<?php

namespace Drupal\codemirror_editor\Plugin\Filter;

use Drupal\Component\Utility\Html;
use Drupal\Core\Form\FormStateInterface;
use Drupal\filter\FilterProcessResult;
use Drupal\filter\Plugin\FilterBase;

/**
 * Provides a 'CodeMirror' filter.
 *
 * @Filter(
 *   id = "codemirror_editor",
 *   title = @Translation("CodeMirror"),
 *   type = Drupal\filter\Plugin\FilterInterface::TYPE_TRANSFORM_IRREVERSIBLE,
 *   settings = {
 *     "lineNumbers" = true,
 *     "foldGutter" = false
 *   }
 * )
 */
class CodeMirrorEditor extends FilterBase {

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {

    $form['lineNumbers'] = [
      '#title' => t('Line numbers'),
      '#type' => 'checkbox',
      '#default_value' => $this->settings['lineNumbers'],
    ];

    $form['foldGutter'] = [
      '#title' => t('Fold gutter'),
      '#type' => 'checkbox',
      '#default_value' => $this->settings['foldGutter'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function process($text, $langcode) {

    $options = $this->settings;
    $options['readOnly'] = TRUE;
    $options['toolbar'] = FALSE;

    $html_dom = Html::load($text);

    $xpath = new \DOMXPath($html_dom);

    $elements = $xpath->query('//code[@data-mode]');

    foreach ($elements as $element) {
      /* @var \DOMElement  $element */
      $mode = $element->getAttribute('data-mode');

      // Support short syntax for modes, i.e. 'html' instead of 'text/html'.
      if (strpos($mode, '/') === FALSE) {
        $mode = self::getModeFromAlias($mode);
      }

      $textarea = $html_dom->createElement('textarea', self::getInnerHtml($element));
      $local_options = ['mode' => $mode] + $options;
      $textarea->setAttribute('data-codemirror', json_encode($local_options));
      // This should slightly reduce flickering while CodeMirror is rendering.
      $textarea->setAttribute('style', 'width: 100%;');

      $element->parentNode->insertBefore($textarea, $element->nextSibling);
      $element->parentNode->removeChild($element);
    }

    $output = Html::serialize($html_dom);
    $output = trim($output);

    if (count($elements)) {
      $build['#attached']['library'][] = 'codemirror_editor/editor';
      \Drupal::service('renderer')->render($build);
    }

    return new FilterProcessResult($output);
  }

  /**
   * {@inheritdoc}
   */
  public function tips($long = FALSE) {
    $tip_arguments = [
      '@expression' => '<code data-mode="mode">...</code>',
    ];
    return $this->t('Syntax highlight code surrounded by the <code>@expression</code> tags.', $tip_arguments);
  }

  /**
   * Converts the DOM element to an HTML snippet.
   *
   * @param \DOMElement $element
   *   The DOM element to serialize.
   *
   * @return string
   *   The HTML code.
   */
  protected static function getInnerHtml(\DOMElement $element) {
    $innerHTML = '';
    foreach ($element->childNodes as $child) {
      $innerHTML .= $element->ownerDocument->saveHTML($child);
    }
    return $innerHTML;
  }

  /**
   * Converts mode alias into mime type suitable for usage in CodeMirror.
   *
   * @param string $alias
   *   The alias of the mime type.
   *
   * @return string|null
   *   A mime type or null if nothing was not found.
   */
  protected static function getModeFromAlias($alias) {
    $modes = codemirror_editor_load_modes();
    codemirror_editor_load_modes();
    foreach ($modes as $mode) {
      foreach ($mode['mime_types'] as $mime_type => $label) {
        $parts = explode('/x-', $mime_type);
        // Some mime types has no 'x-' prefix.
        if (!isset($parts[1])) {
          $parts = explode('/', $mime_type);
        }
        if ($alias == $parts[1]) {
          return $mime_type;
        }
      }
    }
  }

}
